/*-
 * Copyright (C)2024 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#) Copyright (C)2024 @BABOLO http://www.babolo.ru/"
#ident "@(#) $Id: dir.c,v 1.1 2024/11/24 20:57:25 babolo Exp $"

#define BLIN_COMPAT      4
#define MIFE_COMPAT      5

#include <sys/types.h>
#include <sys/event.h>
#include <sys/time.h>
#include <sysexits.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <ufs/ffs/fs.h>
#include <ufs/ufs/dir.h>
#include <babolo/BLINflag.h>
#include <mife.h>
#include "unrmufs2.h"

static void
inout(unrmufs2_cf *cf, struct direct *dentry, uint32_t loc, int t) {
#   define blin_internal_flags (cf->flags & BLIN_MASK)
    if  (BLIN_BIT1 & cf->flags) {
        ifBLIN_QX1( "ino=%u %c%08X  reclen=%u type=%02o namlen=%u=%.*s"
                  , dentry->d_ino
                  , !t ? '@' : '#'
                  , loc
                  , dentry->d_reclen
                  , dentry->d_type
                  , dentry->d_namlen
                  , (int)dentry->d_namlen, dentry->d_name
                  );
    } else {
        printf( "ino=%u %c%08X  reclen=%u type=%02o namlen=%u=%.*s\n"
              , dentry->d_ino
              , !t ? '@' : '#'
              , loc
              , dentry->d_reclen
              , dentry->d_type
              , dentry->d_namlen
              , (int)dentry->d_namlen, dentry->d_name
              );
    }
#   undef blin_internal_flags
}

int
/*****************************************************************************************************
 **                                                                                                 **/
unrmufs2_dir(unrmufs2_cf *cf, struct ufs2_dinode *ino) {                                           /**
 **                                                                                                 **
 *****************************************************************************************************/
#   define blin_internal_flags (cf->flags & BLIN_MASK)
    struct direct      *dentry     ;
    u_int16_t           reclen     ;
    char                buf        [cf->fs->fs_bsize];
    int                 ex         = EX_OK;
    char               *p          ;
    char               *q          ;
    uint                l          ;
    uint8_t             c          ;

    ifBLIN_QX3("+ %"BLIN_X, BLIN_I(ino));
    for (int64_t k = 0; k < UNRMUFS2_NDADDR; k++) {
        if  (!!ino->di_db[k]) {
            if  (UNRMUFS2_RTRAC & cf->flags) {
                ifBLIN_QX0( "%ld Read %06zX(%zu) @%09lX(%lu)@%lX(%lu)"
                          , k
                          , (size_t)cf->fs->fs_bsize
                          , (size_t)cf->fs->fs_bsize
                          , ino->di_db[k] << (DEV_BSHIFT + cf->fs->fs_fsbtodb)
                          , ino->di_db[k] << (DEV_BSHIFT + cf->fs->fs_fsbtodb)
                          , ino->di_db[k]
                          , ino->di_db[k]
                          );
            }
            ex = (int)pread( cf->ifd
                           , buf
                           , (size_t)cf->fs->fs_bsize
                           , ino->di_db[k] << (DEV_BSHIFT + cf->fs->fs_fsbtodb)
                           );
            if  (0 > ex) {
                ifBLIN_QW0( "Can't %ld read %06zX(%zu) @%09lX(%lu)@%lX(%lu)"
                          , k
                          , (size_t)cf->fs->fs_bsize
                          , (size_t)cf->fs->fs_bsize
                          , ino->di_db[k] << (DEV_BSHIFT + cf->fs->fs_fsbtodb)
                          , ino->di_db[k] << (DEV_BSHIFT + cf->fs->fs_fsbtodb)
                          , ino->di_db[k]
                          , ino->di_db[k]
                          );
                continue;
            }
            if (0 < ex) ex = EX_OK;
            p = buf;
            for ( dentry = (struct direct *)p
                ; ino->di_size > (uint64_t)(p - buf)
                ; dentry = (struct direct *)(p += reclen)
                ) {
                reclen = dentry->d_reclen;
                inout(cf, dentry, (uint32_t)(p - buf), 0);
                if  (!reclen) break;
                l = (uint)strnlen(dentry->d_name, dentry->d_namlen);
                if  (20 < ((int)reclen - (int)l)) {
                    l = ((8 + l) & 0xFFFFFFFC) + 4;
                    q = p + l;
                    c = 0;
                    for (uint i = 0; i < (reclen - l); i++) c |= q[i];
                    if  (!!c) {
                        blin_dumb(0, q, reclen - l);
                        for ( dentry = (struct direct *)q
                            ; q < (p + reclen)
                            ; dentry = (struct direct *)(q += dentry->d_reclen)
                            ) {
                            inout(cf, dentry, (uint32_t)(q - buf), 1);
                            if  (!dentry->d_reclen) break;
    }   }   }   }   }   }
    for (int64_t k = 0; k < UNRMUFS2_NIADDR; k++) {
        if  (!!ino->di_ib[k]) {
            if  (UNRMUFS2_RTRAC & cf->flags) {
                ifBLIN_QX0( "%ld Read %06zX(%zu) @%09lX(%lu)@%lX(%lu)"
                          , k
                          , (size_t)cf->fs->fs_bsize
                          , (size_t)cf->fs->fs_bsize
                          , ino->di_ib[k] << DEV_BSHIFT
                          , ino->di_ib[k] << DEV_BSHIFT
                          , ino->di_ib[k]
                          , ino->di_ib[k]
                          );
            }
            if  (0 > pread(cf->ifd, buf, (size_t)cf->fs->fs_bsize, ino->di_ib[k] << DEV_BSHIFT)) {
                ifBLIN_QW0( "Can't %ld read %06zX(%zu) @%09lX(%lu)@%lX(%lu)"
                          , k
                          , (size_t)cf->fs->fs_bsize
                          , (size_t)cf->fs->fs_bsize
                          , ino->di_ib[k] << DEV_BSHIFT
                          , ino->di_ib[k] << DEV_BSHIFT
                          , ino->di_ib[k]
                          , ino->di_ib[k]
                          );
                continue;
    }   }   }
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

int
/*****************************************************************************************************
 **                                                                                                 **/
unrmufs2_findir(unrmufs2_cf *cf, off_t cgoff, uint32_t ndblk) {                                    /**
 **                                                                                                 **
 *****************************************************************************************************/
#   define blin_internal_flags (cf->flags & BLIN_MASK)
    off_t               dgofb  ;
    char                *buf   [cf->fs->fs_fsize];
    int                 ex         = EX_OK;

    ifBLIN_QX3("+ %"BLIN_O"d", cgoff);
    dgofb = (cgoff + ((cf->fs->fs_dblkno - cf->fs->fs_sblkno) << cf->fs->fs_fsbtodb));
    ifBLIN_QX0("%012lX %012lX %012lX", cgoff, dgofb, dgofb + ndblk);
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}
